/*
 *  sicocon.cpp
 *  PhatBoy Control
 *
 *  Created by Arne Scheffler on Sat Feb 22 2003.
 *  Copyright (c) 2003 Arne Scheffler. All rights reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef __sicocon__
#include "sicocon.h"
#endif

#ifndef __plugxmlgui__
#include "plugxmlgui.h"
#endif

struct ParamDesc
{
	const char* name;
	long min;
	long max;
	long value;
};

//------------------------------------------------------------------------
static ParamDesc gParameters[SiCoCon::kNumParameters] =
{
	{"PBCC",			0,	13, 			0},
	{"CtrOUT",			0,	127,			0},
};

//------------------------------------------------------------------------
static const char* mapNames[] =
{
	"1",
	"2",
	"3",
	NULL
};

//------------------------------------------------------------------------
static const char* pbNames[] =
{
	"FILTER CUTOFF",
	"RESONANCE",
	"VIBRATO RATE",
	"VIBRATO DEPTH",
	"VIBRATO DELAY",
	"REVERB",
	"CHORUS",
	"ATTACK",
	"DECAY",
	"SUSTAIN",
	"RELEASE",
	"PAN",
	"VOLUME",
	NULL
};

static CResourceID gEditorResourceID (1, "inspector.xml");

static const char* gEditorTemplateName = "PhatBoyMapperInspectorView";

//------------------------------------------------------------------------
SiCoCon::SiCoCon (const char* uid)
: CMidiEffect (uid)
, playPos (0)
, learn (false)
, paramCCOut (0)
, paramPBCC (0)
{
	flags = kMEHasInspector | kMEIsEditable;
	usage = kMEUsageInsert;
	presetsAreChunks (true);

}

//------------------------------------------------------------------------
SiCoCon::~SiCoCon ()
{
}

//------------------------------------------------------------------------
tresult SiCoCon::initializePlugin (FUnknown* context)
{
	tresult result = CMidiEffect::initializePlugin (context);
	if(result != kResultOk)
		return result;
	
	// create parameter object
	makeParameters ();

	// create UI Factory object
	uiFactory = new CUIFactory (gEditorResourceID, this);
	uiFactory->setTemplateName (CUIFactory::kInspectorTemplate, gEditorTemplateName);
	uiFactory->setTemplateName (CUIFactory::kEditorTemplate, gEditorTemplateName);

	if(accessor)
		accessor->setCanPlayInStop (true);

	return kResultOk;
}

//------------------------------------------------------------------------
tresult SiCoCon::terminatePlugin ()
{
	return CMidiEffect::terminatePlugin ();
}

//------------------------------------------------------------------------
tresult PLUGIN_API SiCoCon::parameterChanged (IParameter* p, long tag)	
{
	if (paramCCOut && paramPBCC)
	{
		switch (tag)
		{
			case kPBCCTag:
			{
				paramCCOut->setValue (map.getOutValue (2, paramPBCC->getValue ()));
				break;
			}
			case kCCOutTag: 
			{
				map.setOutValue (2, paramPBCC->getValue (), paramCCOut->getValue ());
				break;
			}
		}
	}
	return kResultTrue;
}

//------------------------------------------------------------------------
void SiCoCon::makeParameters ()
{
	params = new CPlugParams (this);

	paramPBCC = params->addListParam (kPBCCTag, gParameters[0].name, pbNames);
	paramPBCC->setValue (gParameters[0].value);
	paramCCOut = params->addIntParam (kCCOutTag, gParameters[1].name, gParameters[1].min, gParameters[1].max);
	paramCCOut->setValue (gParameters[1].value);

	paramPBCC->addDependent ((IParameter*)paramCCOut);
}

//------------------------------------------------------------------------
tresult PLUGIN_API SiCoCon::receiveEvent (IMEObjectID event)
{
	if(!accessor)
		return kNotInitialized;

	if (accessor->getStatus (event) == kController)
	{
		if (map.processMidi (event, accessor) == kResultTrue)
			return kResultTrue;
	}
	accessor->passToOutputQueue (event);

	return kResultTrue;
}

//------------------------------------------------------------------------
tresult PLUGIN_API SiCoCon::playAction (long fromPPQ, long toPPQ, bool immediate)
{	
	if(!accessor)
		return kNotInitialized;

	return kResultTrue;
}

//------------------------------------------------------------------------
tresult PLUGIN_API SiCoCon::positAction (long position)
{
	playPos = position;
	return kResultTrue;
}

//------------------------------------------------------------------------
tresult PLUGIN_API SiCoCon::swapAction (long position, long length)
{
	playPos = position;
	return kResultTrue;
}

//------------------------------------------------------------------------
tresult PLUGIN_API SiCoCon::stopAction (long position)
{
	playPos = position;
	return CMidiEffect::stopAction (position);
}

//------------------------------------------------------------------------
tresult PLUGIN_API SiCoCon::startAction (long position)
{
	playPos = position;
	return CMidiEffect::startAction (position);
}

#define CHUNK_SIZE	(3*13*sizeof(char))

//------------------------------------------------------------------------
tresult PLUGIN_API SiCoCon::setChunk (char* chunk, long size)
{
	if (size != CHUNK_SIZE)
		return kResultFalse;

	for (int m = 0; m < 3; m++)
	{
		for (int i = 0; i < 13; i++)
		{
			int value = *chunk++;
			map.setOutValue (m, i, value);
		}
	}
	paramCCOut->setValue (map.getOutValue (2, paramPBCC->getValue ()));
	return kResultTrue;
}

//------------------------------------------------------------------------
tresult PLUGIN_API SiCoCon::getChunk (char* chunk, long* size)
{
	if (chunk)
	{
		for (int m = 0; m < 3; m++)
		{
			for (int i = 0; i < 13; i++)
			{
				*chunk++ = map.getOutValue (m, i);
			}
		}
		return kResultTrue;
	}
	else if (size)
	{
		*size = CHUNK_SIZE;
		return kResultTrue;
	}
	return kResultFalse;
}

static const char map1[] =
{
	32,
	33,
	8,
	9,
	10,
	91,
	93,
	99,
	100,
	5,
	102,
	10,
	7,
};

static const char map2[] =
{
	21,
	22,
	3,
	18,
	2,
	91,
	93,
	11,
	13,
	14,
	15,
	10,
	7,
};

//------------------------------------------------------------------------
// Processing
//------------------------------------------------------------------------
tresult PhatBoyMap::processMap1 (IMEObjectID event, IMEAccessor* accessor)
{
	int byte1 = accessor->getData1 (event);
	int byte2 = accessor->getData2 (event);
	if (mapcc99)
	{
		if (mapcc6)
			return kResultFalse;
		if (mapcc98)
		{
			int data = -1;
			for (char i = 0; i < 13; i++)
			{
				if (map1[i] == mapcc)
				{
					data = ccOut[0][i];
					break;
				}
			}
			if (data == -1) // this should not happen
				return kResultFalse;

			mapcc98 = false;
			mapcc99 = false;
			accessor->setData1 (event, data);
			accessor->passToOutputQueue (event);
			mapcc =0;
			return kResultTrue;
		}
		if (byte1 == 98)
		{
			mapcc = byte2;
			mapcc98 = true;
			return kResultTrue;
		}
	}
	else if (byte1 == 99)
	{
		mapcc99 = true;
		return kResultTrue;
	}
	else if (byte1 == 5)
	{
		accessor->setData1 (event, ccOut[0][9]);
		accessor->passToOutputQueue (event);
		return kResultTrue;
	}
	else if (byte1 == 91)
	{
		accessor->setData1 (event, ccOut[0][5]);
		accessor->passToOutputQueue (event);
		return kResultTrue;
	}
	else if (byte1 == 93)
	{
		accessor->setData1 (event, ccOut[0][6]);
		accessor->passToOutputQueue (event);
		return kResultTrue;
	}
	else if (byte1 == 10)
	{
		accessor->setData1 (event, ccOut[0][11]);
		accessor->passToOutputQueue (event);
		return kResultTrue;
	}
	else if (byte1 == 7)
	{
		accessor->setData1 (event, ccOut[0][12]);
		accessor->passToOutputQueue (event);
		return kResultTrue;
	}
	return kResultFalse;
}

//------------------------------------------------------------------------
tresult PhatBoyMap::processMap2 (IMEObjectID event, IMEAccessor* accessor)
{
	int byte1 = accessor->getData1 (event);
	int byte2 = accessor->getData2 (event);
	if (mapcc99)
	{
		if (mapcc98)
		{
			if (mapcc6)
			{
				int data = -1;
				mapcc98 = false;
				mapcc99 = false;
				mapcc6 = false;
				for (char i = 0; i < 13; i++)
				{
					if (map2[i] == mapcc)
					{
						data = ccOut[1][i];
						break;
					}
				}
				if (data == -1) // this should not happen
					return kResultFalse;

				accessor->setData1 (event, data);
				int data2 = 0; 
				if (mapcc == 21 || mapcc == 22 || mapcc == 3 || mapcc == 18 || mapcc == 14)
					data2 = byte2;
				else
				{
					ccdata = (ccdata - 64) * 16;
					byte2 /= 8;
					data2 = ccdata + byte2;
				}
				accessor->setData2 (event, data2);
				accessor->passToOutputQueue (event);
				mapcc =0;
				return kResultTrue;
			}
			if (byte1 == 6)
			{
				mapcc6 = true;
				ccdata = byte2;
				return kResultTrue;
			}
		}
		if (byte1 == 98)
		{
			mapcc = byte2;
			mapcc98 = true;
			return kResultTrue;
		}
	}
	else if (byte1 == 99)
	{
		mapcc99 = true;
		return kResultTrue;
	}
	else if (byte1 == 5)
	{
		accessor->setData1 (event, ccOut[1][9]);
		accessor->passToOutputQueue (event);
		return kResultTrue;
	}
	else if (byte1 == 91)
	{
		accessor->setData1 (event, ccOut[1][5]);
		accessor->passToOutputQueue (event);
		return kResultTrue;
	}
	else if (byte1 == 93)
	{
		accessor->setData1 (event, ccOut[1][6]);
		accessor->passToOutputQueue (event);
		return kResultTrue;
	}
	else if (byte1 == 10)
	{
		accessor->setData1 (event, ccOut[1][11]);
		accessor->passToOutputQueue (event);
		return kResultTrue;
	}
	else if (byte1 == 7)
	{
		accessor->setData1 (event, ccOut[1][12]);
		accessor->passToOutputQueue (event);
		return kResultTrue;
	}
	return kResultFalse;
}

//------------------------------------------------------------------------
tresult PhatBoyMap::processMap3 (IMEObjectID event, IMEAccessor* accessor)
{
	// this is the easiest map
	int cc = accessor->getData1 (event);
	if (cc > 0 && cc < 14 && ccOut[2][cc-1] != -1)
	{
		accessor->setData1 (event, ccOut [2][cc-1]);
		accessor->passToOutputQueue (event);
		return kResultTrue;
	}
	return kResultFalse;
}

